module TypesWorkshop.Context.Check where

import Data.List (find)
import TypesWorkshop.Context.Data
import TypesWorkshop.ValidationNel

data ResolutionError = ResolutionError { name :: Name }
data TypeError = TypeError { expected, actual :: Type }
               | NameNotFound { notFound :: Name } deriving (Show)

type Context = [Parameter]

resolutionCheck :: Term -> ValidationNel ResolutionError Term
resolutionCheck = resolutionCheckCtx []

resolutionCheckCtx :: Context -> Term -> ValidationNel ResolutionError Term
resolutionCheckCtx ctx term = case term of
    Literal _ -> pure term
    Binary left op right -> 
        Binary <$> resolutionCheckCtx ctx left <*> pure op <*> resolutionCheckCtx ctx right
    Variable v
       | v `notElem` (map paramName ctx) -> problem $ ResolutionError v
       | otherwise  -> pure term
    Lambda params body -> do
        Lambda params <$> resolutionCheckCtx (params ++ ctx) body

shouldBe :: Type -> Type -> ValidationNel TypeError ()
shouldBe actual expected
  | actual == expected = pure ()
  | otherwise          = problem $ TypeError expected actual

sameTypeOperation :: Type -> Type -> Type -> ValidationNel TypeError Type
sameTypeOperation tyLeft tyRight expected = do
    tyLeft `shouldBe` expected
    tyRight `shouldBe` expected
    pure expected

lookupType :: Context -> Name -> Maybe Type
lookupType ctx v = paramType <$> find ((==) v. paramName) ctx

typeCheck :: Term -> ValidationNel TypeError Type
typeCheck = typeCheckCtx []

typeCheckCtx :: Context -> Term -> ValidationNel TypeError Type
typeCheckCtx _ctx (Literal (LitInteger _)) = pure TyInteger
typeCheckCtx _ctx (Literal (LitBoolean _)) = pure TyBoolean
typeCheckCtx ctx (Binary left op right) = do
    tyLeft <- typeCheckCtx ctx left
    tyRight <- typeCheckCtx ctx right
    case op of
        Plus  -> sameTypeOperation tyLeft tyRight TyInteger
        Minus -> sameTypeOperation tyLeft tyRight TyInteger
        And   -> sameTypeOperation tyLeft tyRight TyBoolean
        Or    -> sameTypeOperation tyLeft tyRight TyBoolean
        Equals -> do tyLeft `shouldBe` tyRight
                     pure TyBoolean
typeCheckCtx ctx (Variable v)
  | Just ty <- lookupType ctx v = pure ty
  | otherwise = problem $ NameNotFound v
typeCheckCtx ctx (Lambda params body) =
  TyFunction (map paramType params) <$> typeCheckCtx (params ++ ctx) body