module TypesWorkshop.Generics.Check where

import Control.Monad (forM_)
import Control.Monad.Trans.State
import Data.List (find)
import TypesWorkshop.Generics.Data
import TypesWorkshop.ValidationNel hiding (problem)
import qualified TypesWorkshop.ValidationNel as V
import qualified Unbound.Generics.LocallyNameless as U
import qualified Unbound.Generics.Unify as U

data TypeError = TypeError { expected, actual :: Type }
               | NameNotFound { notFound :: TermVariable } 
               | NoTypeArguments { name :: TermVariable }
               | FunctionExpected
               | NoArguments
               deriving (Show)

type Context = [(TermVariable, TypeScheme)]

newtype TypeCheckM a 
  = TypeCheckM { unTypeCheckM :: U.UnificationMT Type (U.FreshMT (ValidationNel TypeError)) a }
  deriving (Functor, Applicative, Monad, U.Fresh, U.Unification Type)

-- layers and layers
problem :: TypeError -> TypeCheckM a
problem e = TypeCheckM $
            U.UnificationMT $ StateT $ \_ -> 
            U.FreshMT $ StateT $ \_ -> 
            V.problem e

shouldBe :: Type -> Type -> TypeCheckM ()
shouldBe actual expected
  | actual == expected = pure ()
  | otherwise          = problem $ TypeError expected actual

sameTypeOperation :: Type -> Type -> Type -> TypeCheckM Type
sameTypeOperation tyLeft tyRight expected = do
    tyLeft `shouldBe` expected
    tyRight `shouldBe` expected
    pure expected

lookupType :: Context -> TermVariable -> Maybe TypeScheme 
lookupType ctx v = snd <$> find ((==) v. fst) ctx

typeCheck :: Term -> TypeCheckM Type
typeCheck = typeCheckCtx []

headTypeCheckCtx :: Context -> Term -> TypeCheckM TypeScheme
headTypeCheckCtx ctx (Variable v)
  | Just ty <- lookupType ctx v = pure ty
  | otherwise                   = problem $ NameNotFound v
headTypeCheckCtx ctx other = scheme <$> typeCheckCtx ctx other

typeCheckCtx :: Context -> Term -> TypeCheckM Type
typeCheckCtx _ctx (Literal (LitInteger _)) = pure TyInteger
typeCheckCtx _ctx (Literal (LitBoolean _)) = pure TyBoolean
typeCheckCtx ctx (Binary left op right) = do
    tyLeft <- typeCheckCtx ctx left
    tyRight <- typeCheckCtx ctx right
    case op of
        Plus  -> sameTypeOperation tyLeft tyRight TyInteger
        Minus -> sameTypeOperation tyLeft tyRight TyInteger
        And   -> sameTypeOperation tyLeft tyRight TyBoolean
        Or    -> sameTypeOperation tyLeft tyRight TyBoolean
        Equals -> do tyLeft `shouldBe` tyRight
                     pure TyBoolean
typeCheckCtx ctx (Variable v)
  | Just ty <- lookupType ctx v = do
      (tyvars, inner) <- U.unbind ty
      case tyvars of
        [] -> problem $ NoTypeArguments v
        _  -> pure inner
  | otherwise = problem $ NameNotFound v
typeCheckCtx ctx (Lambda lambda) = do
  (params, body) <- U.unbind lambda
  let newContext = map (\p -> (paramName p, scheme $ paramType p)) params ++ ctx
  TyFunction (map paramType params) <$> typeCheckCtx newContext body
typeCheckCtx ctx (Application hd tyargs args) = do
  headTy <- headTypeCheckCtx ctx hd
  let headTy' = U.instantiate headTy tyargs
  case headTy' of
    TyFunction argTy retTy
      | length argTy == length args -> do
          forM_ (zip argTy args) $ \(ty, arg) -> do
            obtainedArgTy <- typeCheckCtx ctx arg
            obtainedArgTy `shouldBe` ty
          pure retTy
      | otherwise -> problem NoArguments
    _ -> problem FunctionExpected
