package com.serranofp.types.context

import arrow.core.EitherNel
import arrow.core.nel
import arrow.core.raise.either
import arrow.core.raise.ensure
import arrow.core.raise.zipOrAccumulate

sealed interface Term

sealed interface Literal : Term {
    data class Integer(val value: Int) : Literal
    data object True : Literal
    data object False : Literal
}

enum class Operation {
    Plus, Minus, And, Or, Equals
}

data class Binary(val left: Term, val operation: Operation, val right: Term) : Term

typealias Name = String
data class Variable(val name: Name) : Term

data class Parameter(val name: Name, val type: Type)
data class Lambda(val parameters: List<Parameter>, val body: Term) : Term

// data class Val(val name: Name, val type: Type, val body: Term) : Term
// data class Application(val callee: Term, val arguments: List<Term>) : Term

data class ResolutionError(val name: Name)
fun Term.resolutionCheck(): EitherNel<ResolutionError, Term> = resolutionCheck(emptyList())
fun Term.resolutionCheck(context: List<Name>): EitherNel<ResolutionError, Term> = either {
    val me = this@resolutionCheck
    when (me) {
        is Literal -> me
        is Binary -> zipOrAccumulate(
            { me.left.resolutionCheck(context).bind() },
            { me.right.resolutionCheck(context).bind() }
        ) { l, r -> Binary(l, me.operation, r) }
        is Variable -> {
            ensure(me.name in context) { ResolutionError(me.name).nel() }
            me
        }
        is Lambda -> {
            val newBody = me.body.resolutionCheck(me.parameters.map { it.name } + context).bind()
            Lambda(me.parameters, newBody)
        }
    }
}