package com.serranofp.types.context

import arrow.core.Either
import arrow.core.raise.either
import arrow.core.raise.ensure

sealed interface Type {
    data object Integer : Type
    data object Boolean : Type
    data class Function(val arguments: List<Type>, val returnType: Type) : Type
}

sealed interface TypeError {
    data class Mismatch(val expected: Type, val actual: Type): TypeError
    data class Resolution(val name: Name): TypeError
}

fun Term.typeCheck(): Either<TypeError, Type> = typeCheck(emptyMap())

fun Term.typeCheck(context: Map<Name, Type>): Either<TypeError, Type> = either {
    // utility function to check types
    infix fun Type.shouldBe(other: Type) {
        ensure(this == other) { TypeError.Mismatch(other, this) }
    }

    when (this@typeCheck) {
        is Literal.Integer -> Type.Integer
        Literal.True, Literal.False -> Type.Boolean
        is Binary -> {
            val leftType = left.typeCheck(context).bind()
            val rightType = right.typeCheck(context).bind()
            when (operation) {
                Operation.Plus, Operation.Minus -> {
                    leftType shouldBe Type.Integer
                    rightType shouldBe Type.Integer
                    Type.Integer
                }
                Operation.And, Operation.Or -> {
                    leftType shouldBe Type.Boolean
                    rightType shouldBe Type.Boolean
                    Type.Boolean
                }
                Operation.Equals -> {
                    rightType shouldBe leftType
                    Type.Boolean
                }
            }
        }
        is Variable -> {
            ensure(name in context) { TypeError.Resolution(name) }
            context.getValue(name)
        }
        is Lambda -> {
            val bodyContext = context + parameters.associate { it.name to it.type }
            val returnType = body.typeCheck(bodyContext).bind()
            Type.Function(parameters.map { it.type }, returnType)
        }
    }
}
