package com.serranofp.types.generics

import arrow.core.EitherNel
import arrow.core.nel
import arrow.core.raise.either
import arrow.core.raise.ensure
import arrow.core.raise.zipOrAccumulate

sealed interface Term

sealed interface Literal : Term {
    data class Integer(val value: Int) : Literal
    data object True : Literal
    data object False : Literal
}

enum class Operation {
    Plus, Minus, And, Or, Equals
}

data class Binary(val left: Term, val operation: Operation, val right: Term) : Term

typealias Name = String
data class Variable(val name: Name) : Term

data class Parameter(val name: Name, val type: Type)
data class Lambda(val parameters: List<Parameter>, val body: Term) : Term

data class Application(
    val callee: Term,
    val typeArguments: List<Type>,
    val valueArguments: List<Term>
) : Term

data class ResolutionError(val name: Name)
fun Term.resolutionCheck(): EitherNel<ResolutionError, Term> = resolutionCheck(emptyList(), emptyList())
fun Term.resolutionCheck(typeContext: List<Name>, valueContext: List<Name>): EitherNel<ResolutionError, Term> = either {
    val me = this@resolutionCheck
    when (me) {
        is Literal -> me
        is Binary -> zipOrAccumulate(
            { me.left.resolutionCheck(typeContext, valueContext).bind() },
            { me.right.resolutionCheck(typeContext, valueContext).bind() }
        ) { l, r -> Binary(l, me.operation, r) }
        is Variable -> {
            ensure(me.name in valueContext) { ResolutionError(me.name).nel() }
            me
        }
        is Lambda -> {
            val newBody = me.body.resolutionCheck(
                typeContext,
                me.parameters.map { it.name } + valueContext
            ).bind()
            Lambda(me.parameters, newBody)
        }
        is Application -> zipOrAccumulate<_, Term, List<Type>, List<Term>, Term>(
            { me.callee.resolutionCheck(typeContext, valueContext).bind() },
            { me.typeArguments.mapOrAccumulate { it.resolutionCheck(typeContext).bind() } },
            { me.valueArguments.mapOrAccumulate { it.resolutionCheck(typeContext, valueContext).bind() } }
        ) { c, t, v -> Application(c, t, v) }
    }
}