package com.serranofp.types.generics

import arrow.core.EitherNel
import arrow.core.nel
import arrow.core.raise.either
import arrow.core.raise.ensure
import arrow.core.raise.mapOrAccumulate
import arrow.core.raise.zipOrAccumulate
import com.serranofp.types.unification.BasicFunctor
import com.serranofp.types.unification.BasicTerm
import com.serranofp.types.unification.BasicVariable
import java.lang.IllegalArgumentException

sealed interface Type : BasicTerm {
    data class Parameter(val name: String)
            : Type, BasicFunctor("!name", emptyList())
    data class Function(val arguments: List<Type>, val returnType: Type)
            : Type, BasicFunctor("->", arguments + returnType)
    data class Regular(val constructor: TypeHead, val arguments: List<Type>)
            : Type, BasicFunctor("$constructor", arguments)

    data class Variable(val name: String, val unique: Long)
            : Type, BasicVariable("?name?unique")

    companion object {
        val Integer: Type = Regular(TypeHead.Integer, emptyList())
        val Boolean: Type = Regular(TypeHead.Boolean, emptyList())
    }
}

fun Type.scheme(): TypeScheme = TypeScheme(emptyList(), this)
fun Type.substitute(parameter: Type, type: Type): Type {
    if (parameter !is Type.Parameter && parameter !is Type.Variable)
        throw IllegalArgumentException()

    fun Type.go(): Type = this.substitute(parameter, type)
    return when (this) {
        is Type.Parameter -> this
        is Type.Function -> Type.Function(arguments.map { it.go() }, returnType.go())
        is Type.Regular -> Type.Regular(constructor, arguments.map { it.go() })
        is Type.Variable -> if (this == parameter) type else this
    }
}

sealed interface TypeHead {
    data object Integer : TypeHead
    data object Boolean : TypeHead
    data class Other(val name: Name): TypeHead
}

data class TypeScheme(
    val typeParameters: List<Name>,
    val type: Type
)

fun TypeScheme.instantiate(types: List<Type>): Type {
    require(types.size == this.typeParameters.size)
    val (freshTypeParams, freshType) = Freshener.fresh(this)
    var current = freshType
    for ((param, type) in freshTypeParams.zip(types)) {
        current = current.substitute(param, type)
    }
    return current
}

fun TypeScheme.resolutionCheck(): EitherNel<ResolutionError, TypeScheme> = either {
    TypeScheme(typeParameters, type.resolutionCheck(typeParameters).bind())
}

fun Type.resolutionCheck(context: List<Name>): EitherNel<ResolutionError, Type> = either {
    val me = this@resolutionCheck
    when (me) {
        is Type.Function -> zipOrAccumulate(
            { me.arguments.mapOrAccumulate { it.resolutionCheck(context).bind() } },
            { me.returnType.resolutionCheck(context).bind() }
        ) { a, r -> Type.Function(a, r) }
        is Type.Regular -> Type.Regular(
            me.constructor,
            mapOrAccumulate(me.arguments) { it.resolutionCheck(context).bind() }
        )
        is Type.Parameter -> {
            ensure(me.name in context) { ResolutionError(me.name).nel() }
            me
        }
        is Type.Variable -> {
            // type variables do not arise during resolution
            me
        }
    }
}

typealias Context = Map<Name, TypeScheme>

sealed interface TypeError {
    data class Mismatch(val expected: Type, val actual: Type): TypeError
    data class FunctionExpected(val actual: Type): TypeError
    data class Resolution(val name: Name): TypeError
    data object WrongNumberOfTypeArguments: TypeError
    data object WrongNumberOfArguments: TypeError
}
