package com.serranofp.types.generics

import arrow.core.EitherNel
import arrow.core.nel
import arrow.core.raise.either
import arrow.core.raise.ensure
import arrow.core.raise.mapOrAccumulate

fun Term.typeCheck(
    typeContext: List<Name>,
    valueContext: Map<Name, TypeScheme>
): EitherNel<TypeError, Type> = either {
    // utility function to check types
    infix fun Type.shouldBe(other: Type) {
        ensure(this == other) { TypeError.Mismatch(other, this).nel() }
    }

    when (this@typeCheck) {
        is Literal.Integer -> Type.Integer
        Literal.True, Literal.False -> Type.Boolean
        is Binary -> {
            val leftType = left.typeCheck(typeContext, valueContext).bind()
            val rightType = right.typeCheck(typeContext, valueContext).bind()
            when (operation) {
                Operation.Plus, Operation.Minus -> {
                    leftType shouldBe Type.Integer
                    rightType shouldBe Type.Integer
                    Type.Integer
                }
                Operation.And, Operation.Or -> {
                    leftType shouldBe Type.Boolean
                    rightType shouldBe Type.Boolean
                    Type.Boolean
                }
                Operation.Equals -> {
                    rightType shouldBe leftType
                    Type.Boolean
                }
            }
        }
        is Application -> {
            val headType = callee.headType(typeContext, valueContext).bind()
            ensure(headType.typeParameters.size == typeArguments.size) {
                TypeError.WrongNumberOfTypeArguments.nel()
            }
            val functionType = headType.instantiate(typeArguments)
            ensure(functionType is Type.Function) {
                TypeError.FunctionExpected(functionType).nel()
            }
            ensure(functionType.arguments.size == valueArguments.size) {
                TypeError.WrongNumberOfArguments.nel()
            }
            mapOrAccumulate(functionType.arguments.zip(valueArguments)) { (expectedType, arg) ->
                val argumentType = arg.typeCheck(typeContext, valueContext).bind()
                argumentType shouldBe expectedType
            }
            functionType.returnType
        }
        is Variable -> {
            val type = this@typeCheck.headType(typeContext, valueContext).bind()
            ensure(type.typeParameters.isEmpty()) { TypeError.WrongNumberOfTypeArguments.nel() }
            type.type
        }
        is Lambda -> {
            val bodyContext = valueContext + parameters.associate { it.name to it.type.scheme() }
            val returnType = body.typeCheck(typeContext, bodyContext).bind()
            Type.Function(parameters.map { it.type }, returnType)
        }
    }
}

fun Term.headType(
    typeContext: List<Name>,
    valueContext: Map<Name, TypeScheme>
): EitherNel<TypeError, TypeScheme> = either {
    when (this@headType) {
        is Variable -> {
            ensure(name in valueContext) { TypeError.Resolution(name).nel() }
            valueContext.getValue(name)
        }
        else -> this@headType.typeCheck(typeContext, valueContext).bind().scheme()
    }
}