package com.serranofp.types.unification

import arrow.core.raise.Raise

/**
 * Unifiable terms with heads taken from [H]
 */
interface UnifiableTerm<out H, V: Comparable<V>, T: UnifiableTerm<H, V, T>> {
    val variable: V?
    fun split(): Pair<H, List<T>>?
    operator fun contains(variable: V): Boolean
    fun substitute(substitution: Substitution<V, T>): T
}

abstract class UnifiableVariable<H, V: Comparable<V>, T: UnifiableTerm<H, V, T>>: UnifiableTerm<H, V, T> {
    abstract override val variable: V
    override fun split(): Pair<H, List<T>>? = null
    override operator fun contains(variable: V): Boolean = variable == this.variable
    override fun substitute(substitution: Substitution<V, T>): T = when (val result = substitution[variable]) {
        null -> variable as T
        else -> result
    }
    override fun toString(): String = "$variable"
}

abstract class UnifiableFunctor<H, V: Comparable<V>, T: UnifiableTerm<H, V, T>>: UnifiableTerm<H, V, T> {
    override val variable: V? = null
    abstract val head: H
    abstract val subterms: List<T>
    override fun split(): Pair<H, List<T>>? = head to subterms
    override fun contains(variable: V): Boolean = subterms.any { variable in it }
    override fun toString(): String = when {
        subterms.isEmpty() -> "$head"
        else -> "$head(${subterms.joinToString()})"
    }
}

typealias Substitution<V, T> = Map<V, T>

enum class UnifierStatus {
    VALID, OCCURS_CHECK, MISMATCH_HEAD, MISMATCH_ARGS
}

fun <E, H, V: Comparable<V>, T: UnifiableTerm<H, V, T>> Raise<E>.unifyOrFail(
    unifier: Unifier<H, V, T>, x: T, y: T, raise: (UnifierStatus) -> E
) {
    unifier.unify(x, y)
    if (unifier.status != UnifierStatus.VALID)
        raise(raise(unifier.status))
}

class Unifier<H, V: Comparable<V>, T: UnifiableTerm<H, V, T>> {
    var status: UnifierStatus = UnifierStatus.VALID

    // invariant: substitution from smaller to larger variables
    val substitution: MutableMap<V, T> = mutableMapOf()

    fun unify(x: T, y: T) {
        val xVar = x.variable
        val yVar = y.variable
        when {
            xVar != null && xVar == yVar -> return // if they are the same variable, we are OK
            yVar != null && (xVar == null || yVar < xVar) -> unify(y, x) // put the smallest variable always at the front
            xVar != null -> unifyVar(xVar, y) // variable := term
            else -> unifyTerm(x, y)
        }
    }

    private fun unifyVar(xVar: V, y: T) {
        // occurs check
        if (xVar in y) {
            status = UnifierStatus.OCCURS_CHECK
            return
        }
        // do we have a previous substitution for x?
        when (val previousX = substitution[xVar]) {
            null -> {
                // apply to the current substitution
                for ((v, t) in substitution) { substitution[v] = t.substitute(mapOf(xVar to y)) }
                // record the new substitution
                substitution[xVar] = y
            }
            else -> unify(previousX, y)
        }
    }

    private fun unifyTerm(x: T, y: T) {
        val (xHead, xTerms) = x.split()!!
        val (yHead, yTerms) = y.split()!!
        // different heads
        if (xHead != yHead) {
            status = UnifierStatus.MISMATCH_HEAD
            return
        }
        // different # of arguments
        if (xTerms.size != yTerms.size) {
            status = UnifierStatus.MISMATCH_ARGS
            return
        }
        // correct case, unify recursively
        xTerms.zip(yTerms, ::unify)
    }
}